/*
 * Routines to provide ordered writing to stdout or to a file
 *
 * This first version simply forwards data to a single process that
 * then outputs the data.
 *
 * For output to a file that this software opens (rather than stdout, for
 * example), it could use parallel I/O, though the file sizes should rarely
 * need that.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "mpi.h"

typedef struct TextFileData *TextFile;

int TextFileWriteOpen(MPI_Comm comm, int root, const char *fname, MPI_Info info,
		      TextFile *fh_p);
int TextFileWriteOpenFromFH(MPI_Comm comm, int root, FILE *fp, MPI_Info info,
			    TextFile *fh);
int TextFileWriteOrdered(TextFile fh, const char *buf, size_t len);
int TextFileWritePrintf(TextFile fh, const char *restrict format, ...); 
int TextFileWriteFlush(TextFile fh);
int TextFileWriteClose(TextFile fh);
int TextFileWriteFree(TextFile *fh_p);

/* -- end of public header -- */

#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

typedef struct TextFileData {
    MPI_Comm comm;
    int      root, commrank, commsize;
    char     *lbuf;                    /* Local buffer */
    int      lbufAllocLen;
    int      lbufLen;
    char     *wbuf;                    /* Buffer used to accumulate writes, */
    int      wbufAllocLen;             /* valid on root only */
    int      wbufLen;
    int      fd;                       /* Unix for for output; valid at root */
} TextFileData;

static int wbufAllocLenDefault = 65536;
static int lbufAllocLenDefault = 4096;

int TextFileiInitFH(MPI_Comm comm, int root, MPI_Info info, TextFile *fh_p);

int TextFileWriteOpen(MPI_Comm comm, int root, const char *fname, MPI_Info info,
		      TextFile *fh_p)
{
    TextFile     fh;
    int          err;

    err = TextFileiInitFH(comm, root, info, fh_p);
    if (err) return err;

    /* Specific to each open style */
    fh = *fh_p;
    if (fh->commrank == fh->root) {
	/* Decide how to handle mask.  fopen uses the values below (0666) */
	fh->fd = open(fname, O_WRONLY | O_CREAT,
	      S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
	if (fh->fd < 0) {
	    return MPI_ERR_OTHER;
	}
    }

    *fh_p = fh;
    return MPI_SUCCESS;
}
int TextFileWriteOpenFromFH(MPI_Comm comm, int root, FILE *fp, MPI_Info info,
			    TextFile *fh_p)
{
    TextFile     fh;
    int          err;

    err = TextFileiInitFH(comm, root, info, fh_p);
    if (err) return err;

    /* Specific to each open style */
    fh = *fh_p;
    if (fh->commrank == root) {
	fh->fd = fileno(fp);
    }

    return MPI_SUCCESS;
}
int TextFileWriteOrdered(TextFile fh, const char *buf, size_t len)
{
    if (len > fh->lbufAllocLen - fh->lbufLen) {
	int reqlen = len + fh->lbufLen;
	fh->lbuf = (char *)realloc(fh->lbuf, reqlen);
	if (!fh->lbuf) return MPI_ERR_NO_MEM;
	fh->lbufAllocLen = reqlen;
    }
    memcpy(fh->lbuf+fh->lbufLen, buf, len);
    fh->lbufLen += len;
    TextFileWriteFlush(fh);

    return MPI_SUCCESS;
}
int TextFileWritePrintf(TextFile fh, const char *restrict format, ...)
{
    int n, availLen;;
    va_list ap;

    availLen = fh->lbufAllocLen - fh->lbufLen;
    va_start(ap, format);
    n = vsnprintf(fh->lbuf+fh->lbufLen, availLen, format, ap);
    if (n > availLen) {
	int len = n + fh->lbufLen;
	fh->lbuf = (char *)realloc(fh->lbuf, n + fh->lbufLen);
	if (!fh->lbuf) return MPI_ERR_NO_MEM;
	fh->lbufAllocLen = len;
	availLen = n;
	n = vsnprintf(fh->lbuf+fh->lbufLen, availLen, format, ap);
    }
    fh->lbufLen += n;
    va_end(ap);

    return MPI_SUCCESS;
}

int TextFileWriteFlush(TextFile fh)
{
    if (fh->commrank == fh->root) {
	int i;
	/* Eventually, can buffer output into wbuf and write only
	   when full, using write */
	for (i=0; i<fh->commsize; i++) {
	    if (i == fh->root) {
		write(fh->fd, fh->lbuf, fh->lbufLen);
		fh->lbufLen = 0;
	    }
	    else {
		MPI_Status st;
		int        len;
		MPI_Probe(i, 0, fh->comm, &st);
		MPI_Get_count(&st, MPI_CHAR, &len);
		if (len < fh->wbufAllocLen) {
		    fh->wbuf         = (char *)realloc(fh->wbuf, len);
		    fh->wbufAllocLen = len;
		}
		MPI_Recv(fh->wbuf, len, MPI_CHAR, i, 0, fh->comm,
			 MPI_STATUS_IGNORE);
		write(fh->fd, fh->wbuf, len);
		fh->wbufLen = 0;
	    }
	}
}
    else {
	/* Use synchronous send to avoid flooding root process.  This
	   relies on the MPI implementation using rendezvous for this,
	   which is the common choice */
	MPI_Ssend(fh->lbuf, fh->lbufLen, MPI_CHAR, fh->root, 0, fh->comm);
    }

    return MPI_SUCCESS;
}

int TextFileWriteClose(TextFile fh)
{
    if (fh->root == fh->commrank) {
	close(fh->fd);
    }
    return MPI_SUCCESS;
}

int TextFileWriteFree(TextFile *fh_p)
{
    TextFile fh = *fh_p;
    free(fh->lbuf);
    if (fh->wbuf) free(fh->wbuf);
    free(fh);
    *fh_p = 0;

    return MPI_SUCCESS;
}

/*
 * Common code to setup the TextFile for Writes
 */
int TextFileiInitFH(MPI_Comm comm, int root, MPI_Info info, TextFile *fh_p)
{
    TextFile     fh = malloc(sizeof(TextFileData));

    if (!fh) {
	/* Panic */
	return MPI_ERR_NO_MEM;
    }
    MPI_Comm_dup(comm, &fh->comm);
    MPI_Comm_rank(fh->comm, &fh->commrank);
    MPI_Comm_size(fh->comm, &fh->commsize);
    fh->root         = root;

    if (fh->commrank == fh->root) {
	fh->wbufAllocLen = wbufAllocLenDefault;
	fh->wbuf         = (char *)malloc(fh->wbufAllocLen);
	fh->wbufLen      = 0;
	if (!fh->wbuf) return MPI_ERR_NO_MEM;
    }
    else {
	fh->wbuf = 0;
    }
    fh->lbufAllocLen = wbufAllocLenDefault;
    fh->lbuf         = (char *)malloc(fh->wbufAllocLen);
    fh->lbufLen      = 0;
    if (!fh->lbuf) return MPI_ERR_NO_MEM;

    fh->fd = -1;

    *fh_p = fh;

    return MPI_SUCCESS;
}


/* Test program */
int main(int argc, char *argv[])
{
    int wsize, wrank;
    TextFile fh;
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &wrank);
    MPI_Comm_size(MPI_COMM_WORLD, &wsize);

    TextFileWriteOpen(MPI_COMM_WORLD, 0, "testfile", MPI_INFO_NULL, &fh);
    TextFileWritePrintf(fh, "I am %d of %d\n", wrank, wsize);
    if (wrank == 2)
	TextFileWritePrintf(fh, "Rank 2 has an extra line of output\n");
    TextFileWriteFlush(fh);
    TextFileWriteClose(fh);
    TextFileWriteFree(&fh);

    MPI_Finalize();
    return 0;
}
